#!/bin/env python3
import argparse
import logging
import re
import sys

from redmine_gitlab_migrator.redmine import RedmineProject, RedmineClient
from redmine_gitlab_migrator.gitlab import GitlabProject, GitlabClient
from redmine_gitlab_migrator.converters import convert_issue, convert_version, load_user_dict
from redmine_gitlab_migrator.logger import setup_module_logging
from redmine_gitlab_migrator.wiki import TextileConverter, WikiPageConverter
from redmine_gitlab_migrator import sql


"""Migration commands for issues and roadmaps from redmine to gitlab
"""

log = logging.getLogger(__name__)

class CommandError(Exception):
    """ An error that will nicely pop up to user and stops program
    """
    def __init__(self, msg):
        self.msg = msg


def parse_args():
    parser = argparse.ArgumentParser(description=__doc__)

    subparsers = parser.add_subparsers(dest='command')

    parser_issues = subparsers.add_parser(
        'issues', help=perform_migrate_issues.__doc__)
    parser_issues.set_defaults(func=perform_migrate_issues)

    parser_pages = subparsers.add_parser(
        'pages', help=perform_migrate_pages.__doc__)
    parser_pages.set_defaults(func=perform_migrate_pages)

    parser_roadmap = subparsers.add_parser(
        'roadmap', help=perform_migrate_roadmap.__doc__)
    parser_roadmap.set_defaults(func=perform_migrate_roadmap)

    parser_redirect = subparsers.add_parser(
        'redirect', help=perform_redirect.__doc__)
    parser_redirect.set_defaults(func=perform_redirect)

    parser_iid = subparsers.add_parser(
        'iid', help=perform_migrate_iid.__doc__)
    parser_iid.set_defaults(func=perform_migrate_iid)

    delete_issues = subparsers.add_parser(
        'delete-issues', help=perform_delete_issues.__doc__)
    delete_issues.set_defaults(func=perform_delete_issues)

    for i in (parser_issues, parser_pages, parser_roadmap, parser_redirect):
        i.add_argument('redmine_project_url')
        i.add_argument(
            '--redmine-key',
            required=True,
            help="Redmine administrator API key")

    for i in (parser_issues, parser_roadmap, parser_iid, parser_redirect, delete_issues):
        i.add_argument('gitlab_project_url')
        i.add_argument(
            '--gitlab-key',
            required=True,
            help="Gitlab administrator API key")

    for i in (parser_issues, parser_pages, parser_roadmap, parser_iid, parser_redirect, delete_issues):
        i.add_argument(
            '--check',
            required=False, action='store_true', default=False,
            help="do not perform any action, just check everything is ready")

        i.add_argument(
            '--debug',
            required=False, action='store_true', default=False,
            help="More output")

        i.add_argument(
            '--no-verify',
            required=False, action='store_false', default=True,
            help="disable SSL certificate verification")

    parser_issues.add_argument(
        '--closed-states',
        required=False,
        help="comma seperated list of redmine states that close an issue, default closed,rejected")

    parser_issues.add_argument(
        '--custom-fields',
        required=False,
        help="comma seperated list of redmine custom filds to migrate")

    parser_issues.add_argument(
        '--user-dict',
        required=False,
        help="file path with redmine user mapping to gitlab user, in YAML format")

    parser_issues.add_argument(
        '--project-members-only',
        required=False, action='store_true', default=False,
        help="get project members instead of all users, useful for gitlab.com")

    parser_issues.add_argument(
        '--keep-id',
        required=False, action='store_true', default=False,
        help="create and delete empty issues for gaps, useful when no ssh is possible (e.g. gitlab.com)")

    parser_issues.add_argument(
        '--keep-title',
        required=False, action='store_true', default=False,
        help="migrate issues with same title, useful when no ssh is possible (e.g. gitlab.com) and don't need to keep id (faster)")

    parser_issues.add_argument(
        '--initial-id',
        required=False,
        help="Initial issue ID, to skip some issues")

    parser_issues.add_argument(
        '--no-sudo', dest='sudo',
        action='store_false',
        default=True,
        help="do not use sudo, use if user is not admin (e.g. gitlab.com)")

    parser_issues.add_argument(
        '--archive-account', dest='archive_acc',
        required=False,
        default=None,
        help="if account doesn't exists in GitLab use this account as default")

    parser_pages.add_argument(
        '--gitlab-wiki',
        required=True,
        help="Path to local cloned copy of the GitLab Wiki's git repository")

    parser_pages.add_argument(
        '--no-history',
        action='store_true',
        default=False,
        help="do not convert the history")

    return parser.parse_args()

def check(func, message, redmine_project, gitlab_project):
    log.info('{}...'.format(message))
    ret = func(redmine_project, gitlab_project)
    if ret:
        log.info('{}... OK'.format(message))
    else:
        log.error('{}... FAILED'.format(message))
        exit(1)

def check_users(redmine_project, gitlab_project):

    redmine_users = redmine_project.get_participants()
    gitlab_users = gitlab_project.get_instance().get_all_users()

    # Filter out anonymous user
    redmine_user_names = [i['login'] for i in redmine_users if i['login'] != '']
    gitlab_user_names = set([i['username'] for i in gitlab_users])

    log.info('Redmine users are: {}'.format(', '.join(redmine_user_names) + ' '))
    log.info('GitLab users are: {}'.format(', '.join(gitlab_user_names) + ' '))

    return gitlab_project.get_instance().check_users_exist(redmine_user_names)

def check_no_issue(redmine_project, gitlab_project):
    return len(gitlab_project.get_issues()) == 0


# check_no_milestone no longer required (milestone will only be created if not exist)
#def check_no_milestone(redmine_project, gitlab_project):
#    return len(gitlab_project.get_milestones()) == 0


def check_origin_milestone(redmine_project, gitlab_project):
    return len(redmine_project.get_versions()) > 0

def perform_migrate_pages(args):
    redmine = RedmineClient(args.redmine_key, args.no_verify)
    redmine_project = RedmineProject(args.redmine_project_url, redmine)

    # Get copy of GitLab wiki repository
    wiki = WikiPageConverter(args.gitlab_wiki)

    # convert all pages including history
    pages = []
    for page in redmine_project.get_all_pages():
        print("Collecting " + page["title"])
        start_version = page["version"] if args.no_history else 1
        for version in range(start_version, page["version"]+1):
            try:
                full_page = redmine_project.get_page(page["title"], version)
                pages.append(full_page)
            except:
                log.error("Error when retrieving " + page["title"] + ", version " + str(version))

    # sort everything by date and convert
    pages.sort(key=lambda page: page["updated_on"])

    for page in pages:
        wiki.convert(page)

def perform_migrate_issues(args):
    closed_states = []
    if (args.closed_states):
        closed_states = args.closed_states.split(',')

    custom_fields = []
    if (args.custom_fields):
        custom_fields = args.custom_fields.split(',')

    if (args.user_dict is not None):
        load_user_dict(args.user_dict)

    redmine = RedmineClient(args.redmine_key, args.no_verify)
    gitlab = GitlabClient(args.gitlab_key, args.no_verify)

    redmine_project = RedmineProject(args.redmine_project_url, redmine)
    gitlab_project = GitlabProject(args.gitlab_project_url, gitlab)

    gitlab_instance = gitlab_project.get_instance()
    if (args.project_members_only):
        gitlab_users_index = gitlab_project.get_members_index()
    else:
        gitlab_users_index = gitlab_instance.get_users_index()
    redmine_users_index = redmine_project.get_users_index()
    milestones_index = gitlab_project.get_milestones_index()
    textile_converter = TextileConverter()

    log.debug('GitLab milestones are: {}'.format(', '.join(milestones_index) + ' '))
    # get issues
    log.info('Getting redmine issues')
    issues = redmine_project.get_all_issues()
    if args.initial_id:
        issues = [issue for issue in issues if int(args.initial_id) <= issue['id']]

    # convert issues
    log.info('Converting issues')
    issues_data = (
        convert_issue(args.redmine_key,
            i, redmine_users_index, gitlab_users_index, milestones_index, closed_states, custom_fields, textile_converter,
            args.keep_id or args.keep_title, args.sudo, args.archive_acc)
        for i in issues)

    # create issues
    log.info('Creating gitlab issues')
    last_iid = int(args.initial_id or 1) - 1
    for data, meta, redmine_id in issues_data:
        if args.check:
            milestone_id = data.get('milestone_id', None)
            if milestone_id:
                try:
                    gitlab_project.get_milestone_by_id(milestone_id)
                except ValueError:
                    raise CommandError(
                        "issue \"{}\" points to unknown milestone_id \"{}\". "
                        "Check that you already migrated roadmaps".format(
                            data['title'], milestone_id))

            log.info('Would create issue "{}" and {} notes.'.format(
                data['title'],
                len(meta['notes'])))

        else:
            if args.keep_id:
                try:
                    fake_meta = {'uploads': [], 'notes': [], 'must_close': False, 'human_time_estimate': None, 'human_total_time_spent': None}
                    if args.sudo:
                        fake_meta['sudo_user'] = meta['sudo_user']
                    while redmine_id > last_iid + 1:
                        created = gitlab_project.create_issue({'title': 'fake'}, fake_meta, gitlab.get_auth_headers())
                        last_iid = created['iid']
                        gitlab_project.delete_issue(created['iid'])
                        log.info('#{iid} {title}'.format(**created))
                except:
                    log.info('create issue "{}" failed'.format('fake'))
                    raise
            try:
                created = gitlab_project.create_issue(data, meta, gitlab.get_auth_headers())
                last_iid = created['iid']
                log.info('#{iid} {title}'.format(**created))
            except:
                log.info('create issue "{}" failed'.format(data['title']))
                raise

def perform_migrate_iid(args):
    """ Should occur after the issues migration
    """

    # access gitlab database with
    # gitlab-rails dbconsole

    gitlab = GitlabClient(args.gitlab_key, args.no_verify)
    gitlab_project = GitlabProject(args.gitlab_project_url, gitlab)
    gitlab_project_id = gitlab_project.get_id()

    regex_saved_iid = r'-RM-([0-9]+)-MR-(.*)'

    sql_cmd = sql.COUNT_UNMIGRATED_ISSUES.format(
        regex=regex_saved_iid, project_id=gitlab_project_id)

    output = sql.run_query(sql_cmd)

    try:
        m = re.match(r'\s*(\d+)\s*', output, re.DOTALL | re.MULTILINE)
        issues_count = int(m.group(1))
    except (AttributeError, ValueError):
        raise ValueError(
            'Invalid output from postgres command: "{}"'.format(output))

    if issues_count > 0:
        log.info('Ready to recover iid for {} issues.'.format(
            issues_count))
    else:
        log.error(
            "No issue to migrate iid, possible causes: "
            "you already migrated iid or you haven't migrated issues yet.")
        exit(1)

    if not args.check:

        # first we change the iid to large values to prevent a
        # duplicate key value violates unique_constraint "index_issues_on_project_id_and_iid"
        # KEY (project_id, iid)=(37, 83) already exists
        sql_cmd1 = sql.UPDATE_IID_ISSUES.format(
            regex=regex_saved_iid, project_id=gitlab_project_id)
        out1 = sql.run_query(sql_cmd1)

        sql_cmd2 = sql.MIGRATE_IID_ISSUES.format(
            regex=regex_saved_iid, project_id=gitlab_project_id)
        out2 = sql.run_query(sql_cmd2)

        try:
            m = re.match(
                r'\s*(\d+)\s*', output,
                re.DOTALL | re.MULTILINE)
            migrated_count = int(m.group(1))
            log.info('Migrated successfully iid for {} issues'.format(
                migrated_count))
        except (IndexError, AttributeError):
            raise ValueError(
                'Invalid output from postgres command: "{}"'.format(output))

def perform_delete_issues(args):
    """ Delete all issues in the gitlab repo
    """
    gitlab = GitlabClient(args.gitlab_key, args.no_verify)
    gitlab_project = GitlabProject(args.gitlab_project_url, gitlab)

    gitlab_issues = gitlab_project.get_issues()
    log.debug('Got {} issue(s) from gitlab.'.format(len(gitlab_issues)))

    for issue in gitlab_issues:
        log.debug('delete issue {}'.format(issue['id']))
        gitlab_project.delete_issue(issue['iid'])

def perform_migrate_roadmap(args):
    redmine = RedmineClient(args.redmine_key, args.no_verify)
    gitlab = GitlabClient(args.gitlab_key, args.no_verify)

    redmine_project = RedmineProject(args.redmine_project_url, redmine)
    gitlab_project = GitlabProject(args.gitlab_project_url, gitlab)

    checks = [
        #(check_no_milestone, 'Gitlab project has no pre-existing milestone'),
        (check_origin_milestone, 'Redmine project contains versions'),
    ]
    for i in checks:
        check(
            *i, redmine_project=redmine_project,
            gitlab_project=gitlab_project)

    versions = redmine_project.get_versions()
    versions_data = (convert_version(i) for i in versions)

    for data, meta in versions_data:
        if args.check:
            log.info("Would create version {}".format(data))
        else:
            created = gitlab_project.create_milestone(data, meta)
            log.info("Version {}".format(created['title']))

def perform_redirect(args):
    redmine = RedmineClient(args.redmine_key, args.no_verify)
    redmine_project = RedmineProject(args.redmine_project_url, redmine)

    # get issues
    redmine_issues = redmine_project.get_all_issues()

    print('# uncomment next line to enable RewriteEngine')
    print('# RewriteEngine On')
    print('# Redirects from {} to {}'.format(args.redmine_project_url, args.gitlab_project_url))

    for issue in redmine_issues:
        print('RedirectMatch 301 ^/issues/{}$ {}/issues/{}'.format(issue['id'], args.gitlab_project_url, issue['id']))

def main():
    args = parse_args()

    if hasattr(args, 'func'):
        if args.debug:
            loglevel = logging.DEBUG
        else:
            loglevel = logging.INFO

        # Configure global logging
        setup_module_logging('redmine_gitlab_migrator', level=loglevel)
        try:
            args.func(args)

        except CommandError as e:
            log.error(e)
            exit(12)



#!/usr/bin/env python
from __future__ import unicode_literals

import os
from setuptools import setup

try:
    README = open(os.path.join(os.path.dirname(__file__), 'README.md')).read()
except IOError:
    README = ''

setup(
    name='redmine-gitlab-migrator',
    version='1.0.3',
    description='Migrate a redmine project to gitlab',
    long_description=README,
    author='Jocelyn Delalande',
    author_email='jdelalande@oasiswork.fr',
    license='GPL',
    url='https://github.com/redmine-gitlab-migrator/redmine-gitlab-migrator',
    packages=['redmine_gitlab_migrator'],
    install_requires=['pyyaml', 'requests', 'GitPython', 'pypandoc'],
    entry_points={
        'console_scripts': [
            'migrate-rg = redmine_gitlab_migrator.commands:main'
        ]
    },
    test_suite='redmine_gitlab_migrator.tests',
    classifiers=[
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)'
    ]
)



""" Convert Redmine objects to gitlab's
"""

import logging
import yaml

log = logging.getLogger(__name__)
user_dict = None

# Utils

def load_user_dict(path):
    global user_dict
    with open(path, 'r') as stream:
        user_dict = yaml.load(stream, Loader=yaml.SafeLoader)

def redmine_username_to_gitlab_username(redmine_username):
    if user_dict is not None and redmine_username in user_dict:
        return user_dict[redmine_username]
    return redmine_username


def redmine_uid_to_gitlab_user(redmine_id, redmine_user_index, gitlab_user_index):
    redmine_login = redmine_user_index[redmine_id]['login']

    # Check if mapping for this user exists
    redmine_login = redmine_username_to_gitlab_username(redmine_login)

    if not redmine_login in gitlab_user_index:
        redmine_login = 'root'
    return gitlab_user_index[redmine_login]

def convert_attachment(redmine_issue_attachment, redmine_api_key):
    """ Convert a list of redmine attachments to gitlab uploads

    :param redmine_issue_attachment: a dict describing redmine-api-style attachment
    :param redmine_api_key: the redmine api key used for getting the attachment without logging in
    :return: a dict describing the attachment
    """
    uploads = {
        'filename': redmine_issue_attachment['filename'],
        'description': redmine_issue_attachment.get('description'),
        'content_url': '{}?key={}'.format(redmine_issue_attachment['content_url'], redmine_api_key),
        'content_type': redmine_issue_attachment.get('content_type', 'application/octet-stream')
    }

    return uploads


def convert_notes(redmine_issue_journals, redmine_user_index, gitlab_user_index, textile_converter, sudo, archive_acc):
    """ Convert a list of redmine journal entries to gitlab notes

    Filters out the empty notes (ex: bare status change)
    Adds metadata as comment

    :param redmine_issue_journals: list of redmine "journals"
    :return: yielded couple ``data``, ``meta``. ``data`` is the API payload for
        an issue note and meta a dict (containing, at the moment, only a
        "sudo_user" key).
    """

    for entry in redmine_issue_journals:
        journal_notes = entry.get('notes', '')
        if journal_notes is not None and len(journal_notes) > 0:
            journal_notes = textile_converter.convert(journal_notes)
            try:
                author = redmine_uid_to_gitlab_user(
                    entry['user']['id'], redmine_user_index, gitlab_user_index)['username']
            except KeyError:
                # In some cases you have anonymous notes, which do not exist in
                # gitlab.
                if archive_acc is not None:
                    author = archive_acc
                    journal_notes = "Archive from user: {}\n\n{}".format(entry['user']['name'], journal_notes)
                else:
                    log.warning(
                        'Redmine user {} is unknown, attribute note '
                        'to current admin\n'.format(entry['user']))
                    author = None
            if not sudo and author is not None:
                creator_text = " by {}".format(author)
            else:
                creator_text = ''
            body = "{}\n\n*(from redmine: written on {}{})*".format(
                journal_notes, entry['created_on'][:10], creator_text)
            data = {'body': body, 'created_at': entry['created_on']}
            if sudo:
                meta = {'sudo_user': author}
            else:
                meta = {}
            yield (data, meta)


def relations_to_string(relations, children, parent_id, issue_id):
    """ Convert redmine formal relations to some denormalized string

    That's the way gitlab does relations, by "mentioning".

    :param relations: list of issues relations
    :param issue_id: the current issue id
    :return a string listing relations.
    """
    l = []
    for i in relations:
        if issue_id == i['issue_id']:
            other_issue_id = i['issue_to_id']
        else:
            other_issue_id = i['issue_id']
        l.append('  * {} #{}'.format(i['relation_type'], other_issue_id))

    for i in children:
        id = i['id']
        l.append('  * {} #{}'.format('child', id))

    if parent_id > 0:
       l.append('  * {} #{}'.format('parent', parent_id))

    return "\n".join(l)

def changesets_to_string(changesets):
    """ Convert redmine formal changesets to some denormalized string

    :param changesets: list of issues changesets
    :return a string listing changesets.
    """
    l = []
    for i in changesets:
        revision = i['revision']
        committed_on = i['committed_on']
        comments = i['comments']
        try:
            user = i['user']['name']
            by_user_str = ' by {}'.format(user)
        except KeyError:
            by_user_str = ''

        l.append('  * Revision {}{} on {}:\n\n```\n{}\n```\n'.format(
            revision, by_user_str, committed_on, comments))

    return "\n".join(l)

def custom_fields_to_string(custom_fields, custom_fields_include):
    """ Convert redmine custom fields to some denormalized string

    :param custom_fields: list of issues custom_fields
    :return a string listing custom_fileds.
    """
    l = []
    for i in custom_fields:
        name = i['name']

        if name in custom_fields_include and i.get('value'):
            # Name: Value
            l.append('  * {}: {}'.format(name, i['value']))

    return "\n".join(l)

# Convertor

def convert_issue(redmine_api_key, redmine_issue, redmine_user_index, gitlab_user_index,
		  gitlab_milestones_index, closed_states, custom_fields_include, textile_converter, keep_title, sudo, archive_acc):

    issue_state = redmine_issue['status']['name']

    if redmine_issue.get('closed_on', None):
        # quick'n dirty extract date
        close_text = ', closed on {}'.format(redmine_issue['closed_on'][:10])
        closed = True
    elif issue_state.lower() in closed_states:
        close_text = ', closed (state: {})'.format(issue_state)
        closed = True
    else:
        close_text = ''
        closed = False

    relations = redmine_issue.get('relations', [])
    children = redmine_issue.get('children', [])
    parent_id = 0
    if redmine_issue.get('parent', None):
        parent_id = redmine_issue['parent']['id']

    relations_text = relations_to_string(relations, children, parent_id, redmine_issue['id'])
    if len(relations_text) > 0:
        relations_text = "\n* Relations:\n" + relations_text

    changesets = redmine_issue.get('changesets', [])
    changesets_text = changesets_to_string(changesets)
    if len(changesets_text) > 0:
        changesets_text = "\n* Changesets:\n" + changesets_text

    custom_fields = redmine_issue.get('custom_fields', [])
    custom_fields_text = custom_fields_to_string(custom_fields, custom_fields_include)
    if len(custom_fields_text) > 0:
        custom_fields_text = "\n* Custom Fields:\n" + custom_fields_text

    labels = [redmine_issue['tracker']['name']]
    if (redmine_issue.get('category')):
        labels.append(redmine_issue['category']['name'])
    if (redmine_issue.get('status')):
        labels.append(redmine_issue['status']['name'])
    if (redmine_issue.get('priority')):
        labels.append(redmine_issue['priority']['name'])

    attachments = redmine_issue.get('attachments', [])
    due_date = redmine_issue.get('due_date', None)
    if keep_title:
        title = redmine_issue['subject']
    else:
        title = '-RM-{}-MR-{}'.format(redmine_issue['id'], redmine_issue['subject'])

    isFromAnonymous = False
    try:
        author_login = redmine_uid_to_gitlab_user(
            redmine_issue['author']['id'], redmine_user_index, gitlab_user_index)['username']
    except KeyError:
        if archive_acc is not None:
            author_login = archive_acc
        else:    
            log.warning(
                'Redmine issue #{} is anonymous, gitlab issue is attributed '
                'to current admin\n'.format(redmine_issue['id']))
            author_login = None
            isFromAnonymous = True

    if not sudo and author_login is not None:
        creator_text = ' by {}'.format(author_login)
    else:
        creator_text = ''

    description = redmine_issue.get('description', "")
    if archive_acc is not None and archive_acc is author_login:
        if isFromAnonymous is True:
            description = "Archive from Anonymous user \n\n{}".format(description)
        else:
            description = "Archive from user: {}\n\n{}".format(redmine_issue['author']['name'], description)
    converted_description = textile_converter.convert(description)
    if converted_description is False:
        converted_description = description
        log.error("Can't convert Redmine issue {} to markdown formatting! Use clear text!".format(
            redmine_issue['id']
        ))
    data = {
        'title': title,
        'description': '{}\n\n*(from redmine: issue id {}, created on {}{}{})*\n{}{}{}'.format(
            converted_description,
            redmine_issue['id'],
            redmine_issue['created_on'][:10],
            creator_text,
            close_text,
            relations_text,
            changesets_text,
            custom_fields_text
        ),
        'due_date': due_date,
        'created_at': redmine_issue['created_on'],
        'labels': ','.join(labels),
    }

    version = redmine_issue.get('fixed_version', None)
    if version:
        if version['name'] in gitlab_milestones_index:
            data['milestone_id'] = gitlab_milestones_index[version['name']]['id']
        else:
            print("Milestone {} doesn't exists in GitLab Project but exists in Redmine!".format(version['name']))

    estimated_hours = redmine_issue.get('estimated_hours', 0)
    spent_hours = redmine_issue.get('spent_hours', 0)

    meta = {
        'notes': list(convert_notes(redmine_issue['journals'],
                          redmine_user_index, gitlab_user_index, textile_converter, sudo, archive_acc)),
        'must_close': closed,
        'uploads': list(convert_attachment(a, redmine_api_key) for a in attachments),
        "human_time_estimate": estimated_hours,
        "human_total_time_spent": spent_hours
    }
    if sudo:
        meta['sudo_user'] = author_login

    assigned_to = redmine_issue.get('assigned_to', None)
    if assigned_to is not None:
        try:
            data['assignee_id'] = redmine_uid_to_gitlab_user(
                assigned_to['id'], redmine_user_index, gitlab_user_index)['id']
        except KeyError:
            if archive_acc is not None:
                data['assignee_id'] = gitlab_user_index[archive_acc]['id']
            else:
                log.warning(
                    'Redmine issue #{} assignee is anonymous. gitlab assinee is attributed '
                    'to current admin\n'.format(redmine_issue['id']))
            if assigned_to['name']:
                data['labels'] = data['labels'] + ',' + assigned_to['name']

    return data, meta, redmine_issue['id']


def convert_version(redmine_version):
    """ Turns a redmine version into a gitlab milestone

    Do not handle the issues linked to the milestone/version.
    Note that redmine do not expose a due date in API.

    :param redmine_version: a dict describing redmine-api-style version
    :rtype: couple: dict, dict
    :return: a dict describing gitlab-api-style milestone and another for meta
    """
    created_on_string = ""
    if 'created_on' not in redmine_version or redmine_version['created_on'] is None:
        log.warning(
            "Created date for version {} isn't given in redmine!".format(redmine_version['name'])
        )
    else:
        created_on_string = ": created on {}".format(redmine_version['created_on'][:10])
        
    milestone = {
        "title": redmine_version['name'],
        "description": '{}\n\n*(from redmine{})*'.format(
            redmine_version.get('description', ""),
            created_on_string)
    }
    if 'due_date' in redmine_version and redmine_version['due_date'] is not None:
        milestone['due_date'] = redmine_version['due_date'][:10]

    must_close = redmine_version['status'] == 'closed'

    return milestone, {'must_close': must_close}
